//------------------------------------------------------------------
// Create by Robert F. Nee.  Use freely.
//
// Rocketry related modules
//
//------------------------------------------------------------------

//------------------------------------------------------------------
// Tools
//------------------------------------------------------------------

//------------------------------------------------------------------
// translate z height

module jog (h = -0.001) {
	translate ([0, 0, h])
		children ();
};

//------------------------------------------------------------------
// Functions to extract segment height and total height from h vectors
// If zero element is 0 then the vector is assumed to be offsets from zero
// otherwise they are segment heights

module stack (a, s)
{
	// if the segment is zero then the stack offset is always zero
	if (s == 0) {
		children ();
	}
	else {
		// If the first element is nonzero then the elements are segment
		// heights so add them up inductively
		if (a[0] > 0) {
			stack (a, s - 1)
				translate ([0, 0, a[s - 1]])
					children ();
		}
		else {
			translate ([0, 0, a[s]])
				children ();
		}
	}
}

function height (a, s) =
	a[0] > 0 ? a[s] : a[s + 1] - a[s]; 

//-------------------------------------------------------------------------------------
// Shapes
//-------------------------------------------------------------------------------------

// Draw a 2d acute arc
// angle	- acute angle
// radius	- radius of the circle

module arc (angle, radius = 1)
{
	// Points for intersectign polygon
	points = [
		[0, 0],
		[2*radius * cos (angle), 2*radius * sin (angle)],
		[2*radius, 0]
	];

	intersection () {
		circle (radius);
		polygon (points);
	}
}

//-------------------------------------------------------------------------------------
// Draw a hollow tube
// h		- tube height
// d		- tube diameter (base)
// t		- tube wall thickness
// d2		- optional tube diameter top

module tube (h, d, t)
{
	slop = 0.001;

	difference () {
		cylinder (h=h, d=d);
	
		translate ([0, 0, -slop])
			cylinder (h=h + 2*slop, d=d - 2 * t);
	}
}

// ---------------------------------------------------------------------------
// Take first child geometry and etch an inset or outset image of the second child geometry.
// f		- How deep or raised the etching is
//		- f < 1.0 is an inset shell, f > 1.0 is an outset shell

module etch (f)
{
	if (f < 1.0) {
		// Inset
		difference () {
			children (0);
		
			difference () {
				children (1);
		
				scale (f)
					children (0);
			}
		}
	}
	else {
		// outset
		union () {
			children (0);

			intersection () {
				children (1);
	
				scale (f)
					children (0);
			}
		}
	}
}

//------------------------------------------------------------------
// Addachment point on base of nosecone.  Meant to be subtracted out
// of target part. 
// l		- length of cylindrical section.  Used to scale part

module attachment_point (l)
{
	scale (l/12)
		union () {
			translate ([4, 0, 3/2])
				cube ([4, 4, 3], center = true);
			translate ([-4, 0, 3/2])
				cube ([4, 4, 3], center = true);
			translate ([0, 0, 3]) rotate ([0, 90, 0])
				cylinder (h = 12, r = 2, center = true);
		}
}

//------------------------------------------------------------------
// Compute y for a given x for a tangent ogive cone in 2D space.
// r - base radius
// p - phi, radius of circle defining cone arc
// x - distance from base of form (not including shoulder)
// returns y which is the radius of the cone at that point

function tan_ogive_y (r, p, x) = sqrt (max (p * p - x * x, 0)) - (p - r); 

//------------------------------------------------------------------
// Draw a tangent ogive cone
// length		- cone length
// diameter	- base diameter
// slice		- number of slices that make up the cone

module tan_ogive_cone (length, diameter, slice)
{
	// height of each slice
	step = length/slice;

	// radius of base of cone
	base_r = diameter/2;

	// Radius of ogive circle
	phi = (base_r * base_r + length * length) / (2 * base_r);

	for (i = [1 : slice]) {
		assign (x = (i - 1) * step) {
			translate ( [0, 0, x] )
			cylinder (
				step,
				tan_ogive_y (base_r, phi, x),
				tan_ogive_y (base_r, phi, x + step)
			);
		}
	}
}

//------------------------------------------------------------------
// Draw a tangent ogive cone with a spherically blunting "cap"
// length		- cone length of UNCAPPED cone
// diameter	- base diameter
// rn		- Radius of nose sphere cap
// slice		- number of slices that make up the cone

module sb_ogive_cone (length, diameter, rn, slice)
{
	// radius of base of cone
	base_r = diameter/2;

	// Radius of ogive circle
	phi = (base_r * base_r + length * length) / (2 * base_r);

	// center of blunting sphere
	x0 = length - sqrt (pow(phi - rn, 2) - pow (phi - base_r, 2));

	// y of tangency point
	yt = rn * (phi - base_r) / (phi - rn);

	// x of tangency point
	xt = x0 - sqrt (rn * rn - yt * yt);

	// height of each slice taking into account the tangency point
	step = (length - xt)/slice;

	for (i = [1 : slice]) {
		assign (x = (i - 1) * step) {
			translate ( [0, 0, x] )
			cylinder (
				step,
				tan_ogive_y (base_r, phi, x),
				tan_ogive_y (base_r, phi, x + step)
			);
		}
	}

	// Nose sphere cap
	translate ([0, 0, length - x0])
		sphere (r = rn);

	echo ("Actual sb_ogive_cone length = ", length - (x0 -rn));
}

//------------------------------------------------------------------
// Compute y for a given x for a elliptical cone in 2D space.
// l - cone length
// r - base radius
// x - distance from base of form (not including shoulder)
// returns y which is the radius of the cone at that point

function elliptical_y (l, r, x) = r * sqrt (max (1.0 - (x*x)/(l*l), 0));

//------------------------------------------------------------------
// Draw a eliptical cone of given dimensions
// height		- height of cone not including shoulder
// diameter	- diameter at base of cone
// slice		- number of slices that make up the cone

module elliptical_cone (length, diameter, slice)
{
	// height of each slice
	step = length/slice;

	// radius of base of cone
	base_r = diameter/2;

	for (i = [1 : slice]) {
		assign (x = (i - 1) * step) {
			translate ( [0, 0, x] )
			cylinder (
				step,
				elliptical_y (length, base_r, x),
				elliptical_y (length, base_r, x + step)
			);
		}
	}
}

//------------------------------------------------------------------
// Compute y for a given x for a parabolic cone in 2D space.
// l - cone length
// r - base radius
// k - degree of parabola from 0 (cone) to 1 (full parabola)
// x - distance from base of form (not including shoulder)
// returns y which is the radius of the cone at that point

function parabolic_y (l, r, k, x) = r * sqrt (max ((2.0 * (x / l) - k * (x*x)/(l*l)) / (2 - k), 0));

//------------------------------------------------------------------
// Draw a parabolic cone of given dimensions
// height		- height of cone not including shoulder
// diameter	- diameter at base of cone
// k			- degree of parabola from 0 (cone) to 1 (full parabola)
// slice		- number of slices that make up the cone

module parabolic_cone (length, diameter, k, slice)
{
	// height of each slice
	step = length/slice;

	// radius of base of cone
	base_r = diameter/2;

	for (i = [1 : slice]) {
		assign (x = (i - 1) * step) {
			echo (parabolic_y (length, base_r, k, length - x - step));
			translate ( [0, 0, x] )
			cylinder (
				step,
				parabolic_y (length, base_r, k, length - x),
				parabolic_y (length, base_r, k, length - x - step)
			);
		}
	}
}

// ---------------------------------------------------------------------------
// spherically blunted cone

module sb_cone (length, base_radius, tip_radius)
{
	// Compute the angle that the side of the cone makes with the base.
	// This is the sum of two angles, the angle between the base and a radial
	// from the base corner to the origin of the nose sphere and this radial
	// and the side of the cone.  The total angle can then be used to length
	// of the conical section and the radius at the top of this section.
	t = atan ((length - tip_radius) / base_radius) + 
		asin (tip_radius/sqrt ((length-tip_radius)*(length-tip_radius) + base_radius * base_radius));

	// Draw the cone and sphere using the angle t calculated above	
	union () {
		cylinder (length - tip_radius + tip_radius * cos (t), base_radius,
			 tip_radius * sin (t));
		translate ([0,0, length - tip_radius])
		sphere (tip_radius);
	}
}

// Alternate version using a hull

module sb_cone (length, base_radius, tip_radius) {
	hull () {
		cylinder (h = 0.01, r = base_radius);
		translate ([0, 0, length - tip_radius])
		sphere (r = tip_radius);
	}
}

//------------------------------------------------------------------
// Draw a fin
// thickness	- fin thickness
// root		- lenth of root
// sweep		- length of sweep
// tip		- length of tip
// width		- width of fin

module fin (thickness, root, tip, sweep, width)
{
	rotate ([90, 0, 0])
		linear_extrude (height = thickness)
			polygon (	[
				[0, 0],
				[0, root],
				[width, root - sweep],
				[width, root - sweep - tip],
			] );
}


